import inspect
import logging
import re
from functools import partial
from typing import Optional

from amaranth.build import Platform
from amaranth.build.run import LocalBuildProducts

from .cmdrunner import CommandRunner
from .logging import logger, logtime
from .project import Project

__all__ = ["add_arguments"]


def add_arguments(np: Project, parser):
    parser.set_defaults(func=partial(main, np))
    match sorted(t.__name__ for t in np.targets):
        case []:
            raise RuntimeError("no buildable targets defined")
        case [first, *rest]:
            parser.add_argument(
                "-b",
                "--board",
                choices=[first, *rest],
                help="which board to build for",
                required=bool(rest),
                **({"default": first} if not rest else {}),
            )
    parser.add_argument(
        "-p",
        "--program",
        action="store_true",
        help="program the design onto the board after building",
    )
    parser.add_argument(
        "-v",
        "--verilog",
        action="store_true",
        help="output debug Verilog",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="don't use cached synthesis",
    )


def main(np: Project, args):
    logger.info("building %s for %s", np.name, args.board)

    platform = np.target_by_name(args.board)
    design = construct_top(np, platform)

    subdir = type(platform).__name__

    with logtime(logging.DEBUG, "elaboration"):
        plan = platform.prepare(
            design,
            np.name,
            debug_verilog=args.verilog,
            yosys_opts="-g",
        )
    fn = f"{np.name}.il"
    size = len(plan.files[fn])
    logger.debug(f"{fn!r}: {size:,} bytes")

    with logtime(logging.DEBUG, "synthesis/pnr"):
        cr = CommandRunner(force=args.force)
        products = None
        def execute_build():
            nonlocal products
            products = plan.execute_local(f"build/{subdir}")

        # The outf doesn't exist here; it's only used for the digest name basis.
        cr.add_process(execute_build,
            infs=[np.path.build(subdir, fn)],
            outf=np.path.build(subdir, np.name))

        cr.run()
        if products is None:
            # XXX: good lord.
            products = LocalBuildProducts(np.path.build())

    if args.program:
        with logtime(logging.DEBUG, "programming"):
            platform.toolchain_program(products, np.name)

    heading = re.compile(r"^\d+\.\d+\. Printing statistics\.$", flags=re.MULTILINE)
    next_heading = re.compile(r"^\d+\.\d+\. ", flags=re.MULTILINE)
    log_file_between(logging.INFO, np.path.build(subdir, f"{np.name}.rpt"), heading, next_heading)

    logger.info("Device utilisation:")
    heading = re.compile(r"^Info: Device utilisation:$", flags=re.MULTILINE)
    next_heading = re.compile(r"^Info: Placed ", flags=re.MULTILINE)
    log_file_between(
        logging.INFO, np.path.build(subdir, f"{np.name}.tim"), heading, next_heading, prefix="Info: "
    )


def construct_top(np: Project, platform: Platform, **kwargs):
    sig = inspect.signature(np.top)
    if "platform" in sig.parameters:
        kwargs["platform"] = platform
    return np.top(**kwargs)


def log_file_between(
    level: int,
    path: str,
    start: re.Pattern,
    end: re.Pattern,
    *,
    prefix: Optional[str] = None,
):
    with open(path, "r") as f:
        for line in f:
            if start.match(line):
                break
        else:
            return

        for line in f:
            if end.match(line):
                return
            line = line.rstrip()
            if prefix is not None:
                line = line.removeprefix(prefix)
            logger.log(level, line)



import sys
from pathlib import Path
from typing import Optional

from amaranth import Elaboratable
from amaranth.build import Platform

from .cxxrtl_platform import CxxrtlPlatform

__all__ = ["Project"]


class Prop:
    def __init__(
        self,
        name: str,
        *,
        description: str,
        required: bool,
        isinstance: Optional[type] = None,
        issubclass: Optional[type] = None,
        issubclass_list: Optional[type] = None,
    ):
        self.name = name
        self.description = description
        self.required = required
        self.isinstance = isinstance
        self.issubclass = issubclass
        self.issubclass_list = issubclass_list

    def validate(self, project):
        assert not (
            self.issubclass and self.issubclass_list
        ), "may only define one of issubclass and issubclass_list"

        if self.required:
            assert hasattr(project, self.name), (
                f"{project.__module__}.{project.__class__.__qualname__} is missing "
                f"property {self.name!r} ({self.description})"
            )
        elif not hasattr(project, self.name):
            return

        attr = getattr(project, self.name)
        if self.isinstance:
            assert isinstance(attr, self.isinstance), (
                f"{project.__module__}.{project.__class__.__qualname__} property "
                f"{self.name!r} ({self.description}) should an instance of "
                f"{self.isinstance!r}, but is {attr!r}"
            )
        if self.issubclass:
            assert issubclass(attr, self.issubclass), (
                f"{project.__module__}.{project.__class__.__qualname__} property "
                f"{self.name!r} ({self.description}) should be a subclass of "
                f"{self.issubclass!r}, but is {attr!r}"
            )
        if self.issubclass_list:
            assert isinstance(attr, list)
            for elem in attr:
                assert issubclass(elem, self.issubclass_list), (
                    f"{project.__module__}.{project.__class__.__qualname__} property "
                    f"{self.name!r} ({self.description}) should be a list of subclasses of "
                    f"{self.issubclass!r}, but has element {attr!r}"
                )


class Project:
    name: str
    top: type[Elaboratable]
    targets: list[type[Platform]]
    cxxrtl_targets: Optional[list[type[CxxrtlPlatform]]]

    origin: Path

    PROPS = [
        Prop(
            "name",
            description="a keyword-like identifier for the project",
            required=True,
            isinstance=str,
        ),
        Prop(
            "top",
            description="a reference to the default top-level elaboratable to be built",
            required=True,
            issubclass=Elaboratable,
        ),
        Prop(
            "targets",
            description="a list of platform classes the elaboratable is targetted for",
            required=True,
            issubclass_list=Platform,
        ),
        Prop(
            "cxxrtl_targets",
            description="a list of niar.CxxrtlPlatform classes the elaboratable is targetted for",
            required=False,
            issubclass_list=CxxrtlPlatform,
        ),
    ]

    def __init_subclass__(cls):
        # We expect to be called from project-root/module/__init.py__ or similar;
        # cls.origin is project-root. Keep going up until we find pyproject.toml.
        origin = Path(sys._getframe(1).f_code.co_filename).absolute().parent
        while True:
            if (origin / "pyproject.toml").is_file():
                cls.origin = origin
                break
            if not any(origin.parents):
                assert False, "could not find pyproject.toml"
            origin = origin.parent

        extras = cls.__dict__.keys() - {"__module__", "__doc__", "origin"}
        for prop in cls.PROPS:
            prop.validate(cls)
            extras -= {prop.name}
        assert extras == set(), f"unknown project properties: {extras}"

    def target_by_name(self, name: str) -> Platform:
        for t in self.targets:
            if t.__name__ == name:
                return t()
        raise KeyError(f"unknown target {name!r}")

    def cxxrtl_target_by_name(self, name: str) -> CxxrtlPlatform:
        for t in self.cxxrtl_targets or []:
            if t.__name__ == name:
                return t()
        raise KeyError(f"unknown CXXRTL target {name!r}")

    def main(self):
        from . import cli

        cli(self)

    @property
    def path(self):
        return ProjectPath(self)


class ProjectPath:
    def __init__(self, np: Project):
        self.np = np

    def __call__(self, *components):
        return self.np.origin.joinpath(*components)

    def build(self, *components):
        return self("build", *components)



import logging
from contextlib import contextmanager
from datetime import datetime
from typing import Optional

__all__ = ["logger", "logtime"]

logging.basicConfig(
    format="[%(asctime)s] %(name)s: %(levelname)s: %(message)s",
    level=logging.DEBUG,
)

logger = logging.getLogger("niar")


@contextmanager
def logtime(level: int, activity: str, /, fail_level: Optional[int] = None):
    global logger
    start = datetime.now()
    logger.log(level, "starting %s", activity)
    try:
        yield
    except:
        finish = datetime.now()
        logger.log(fail_level or level, "%s failed in %s", activity, finish - start)
        raise
    else:
        finish = datetime.now()
        logger.log(level, "%s finished in %s", activity, finish - start)



from abc import ABCMeta, abstractmethod

__all__ = ["CxxrtlPlatform"]


class CxxrtlPlatform(metaclass=ABCMeta):
    default_clk_frequency = property(abstractmethod(lambda _: None))
    uses_zig = False



import argparse
import json
import logging
import os
import shutil
from enum import Enum, nonmember
from functools import partial
from pathlib import Path

from amaranth._toolchain.yosys import find_yosys
from amaranth.back import rtlil

from .build import construct_top
from .cmdrunner import CommandRunner
from .logging import logtime
from .project import Project

__all__ = ["add_arguments"]

CXXFLAGS = [
    "-std=c++17",
    "-g",
    "-pedantic",
    "-Wall",
    "-Wextra",
    "-Wno-zero-length-array",
    "-Wno-unused-parameter",
]


class _Optimize(Enum):
    rtl = "rtl"
    code = "code"
    none = "none"
    all = "all"

    def __str__(self):
        return self.value

    @nonmember
    class ArgparseAction(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            match (values, namespace.optimize):
                case (_Optimize.rtl, _Optimize.code) | (_Optimize.code, _Optimize.rtl):
                    # Choosing rtl or code when the other is chosen becomes all.
                    newval = _Optimize.all
                case _:
                    # Other cases are:
                    #   Choosing all or none always sets it, regardless of old value.
                    #   Choosing rtl or code when old value is all/none/same always sets it.
                    newval = values

            setattr(namespace, self.dest, newval)

    @property
    def opt_rtl(self) -> bool:
        return self in (self.rtl, self.all)

    @property
    def opt_code(self) -> bool:
        return self in (self.code, self.all)


def add_arguments(np: Project, parser):
    parser.set_defaults(func=partial(main, np))
    match sorted(t.__name__ for t in np.cxxrtl_targets or []):
        case []:
            raise RuntimeError("no cxxrtl targets defined")
        case [first, *rest]:
            parser.add_argument(
                "-t",
                "--target",
                choices=[first, *rest],
                help="which CXXRTL target to build",
                required=bool(rest),
                **({"default": first} if not rest else {}),
            )
    parser.add_argument(
        "-c",
        "--compile",
        action="store_true",
        help="compile only; don't run",
    )
    parser.add_argument(
        "-O",
        "--optimize",
        action=_Optimize.ArgparseAction,
        type=_Optimize,
        choices=_Optimize,
        default=_Optimize.all,
        help="build with optimizations — may be specified multiple times. default if unspecified: all.",
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="generate source-level debug information",
    )
    parser.add_argument(
        "-v",
        "--vcd",
        action="store",
        type=str,
        help="output a VCD file",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="don't use cached compilations",
    )


def main(np: Project, args):
    yosys = find_yosys(lambda ver: ver >= (0, 10))

    platform = np.cxxrtl_target_by_name(args.target)
    design = construct_top(np, platform)

    subdir = type(platform).__name__
    os.makedirs(np.path.build(subdir), exist_ok=True)

    cr = CommandRunner(force=args.force)

    with logtime(logging.DEBUG, "elaboration"):
        il_path = np.path.build(subdir, f"{np.name}.il")
        rtlil_text = rtlil.convert(design, name=np.name, platform=platform)
        with open(il_path, "w") as f:
            f.write(rtlil_text)

        cxxrtl_cc_path = np.path.build(subdir, f"{np.name}.cc")
        yosys_script_path = _make_absolute(np.path.build(subdir, f"{np.name}.ys"))
        black_boxes = {}

        with open(yosys_script_path, "w") as f:
            for box_source in black_boxes.values():
                f.write(f"read_rtlil <<rtlil\n{box_source}\nrtlil\n")
            f.write(f"read_rtlil {_make_absolute(il_path)}\n")
            if args.optimize.opt_rtl:
                f.write("opt\n")
                f.write(f"write_rtlil {_make_absolute(il_path)}.opt\n")
            else:
                # Allow apples-to-apples comparison of generated RTLIL by
                # rewriting it with Yosys.
                f.write(f"write_rtlil {_make_absolute(il_path)}.noopt\n")
            f.write(f"write_cxxrtl -header {_make_absolute(cxxrtl_cc_path)}\n")

        def rtlil_to_cc():
            # "opt" without "proc" generates a bunch of warnings like:
            #
            #   Warning: Ignoring module ili9341spi.initter because it contains
            #   processes (run 'proc' command first).
            #
            # Those passes warn and don't run (instead of removing things we
            # care about), so just let them. We don't want to run "proc" --- it
            # dovetails poorly with CXXRTL and the result is slower than had we
            # done nothing at all.  Note the following option taken by
            # "write_cxxrtl":
            #
            #   -noproc
            #
            #       don't convert processes to netlists. in most designs,
            #       converting processes significantly improves evaluation
            #       performance at the cost of slight increase in compilation
            #       time.
            #
            # Presumably "proc" does it worse, for CXXRTL's purposes.
            yosys.run(["-q", yosys_script_path], ignore_warnings=True)

        cr.add_process(rtlil_to_cc,
            infs=[il_path, yosys_script_path],
            outf=cxxrtl_cc_path)
        cr.run()

    with logtime(logging.DEBUG, "compilation"):
        cc_odep_paths = {cxxrtl_cc_path: (np.path.build(subdir, f"{np.name}.o"), [])}
        depfs = list(np.path("cxxrtl").glob("**/*.h"))
        for path in np.path("cxxrtl").glob("**/*.cc"):
            # XXX: we make no effort to distinguish cxxrtl/a.cc and cxxrtl/dir/a.cc.
            cc_odep_paths[path] = (np.path.build(subdir, f"{path.stem}.o"), depfs)

        cxxflags = CXXFLAGS + [
            f"-DCLOCK_HZ={int(platform.default_clk_frequency)}",
            *(["-O3"] if args.optimize.opt_code else ["-O0"]),
            *(["-g"] if args.debug else []),
        ]
        if platform.uses_zig:
            cxxflags += [
                "-DCXXRTL_INCLUDE_CAPI_IMPL",
                "-DCXXRTL_INCLUDE_VCD_CAPI_IMPL",
            ]

        for cc_path, (o_path, dep_paths) in cc_odep_paths.items():
            cmd = [
                "c++",
                *cxxflags,
                f"-I{np.path.build(subdir)}",
                f"-I{yosys.data_dir() / "include" / "backends" / "cxxrtl" / "runtime"}",
                "-c",
                cc_path,
                "-o",
                o_path,
            ]
            if platform.uses_zig:
                cmd = ["zig"] + cmd
            cr.add_process(cmd, infs=[cc_path] + dep_paths, outf=o_path)

        # Not feasible to do these per-CXXRTL platform, as clangd won't find
        # them (and how could it know which to choose?). Worth noting it checks
        # inside a directory named "build" specifically.
        with open(np.path.build("compile_commands.json"), "w") as f:
            json.dump(
                [{
                    "directory": str(np.path()),
                    "file": file,
                    "arguments": arguments,
                } for file, arguments in cr.compile_commands.items()],
                f,
            )

        cr.run()

        exe_o_path = np.path.build(subdir, np.name)
        cc_o_paths = [o_path for (o_path, _) in cc_odep_paths.values()]
        if platform.uses_zig:
            # Note that we don't clear zig's cache on args.force.
            cmd = [
                "zig",
                "build",
                f"-Dclock_hz={int(platform.default_clk_frequency)}",
                f"-Dyosys_data_dir={yosys.data_dir()}",
            ] + [
                # Zig really wants relative paths.
                f"-Dcxxrtl_o_path=../{p.relative_to(np.path())}" for p in cc_o_paths
            ]
            if args.optimize.opt_code:
                cmd += ["-Doptimize=ReleaseFast"]
            outf = "cxxrtl/zig-out/bin/cxxrtl"
            cr.add_process(cmd,
                infs=cc_o_paths + list(np.path("cxxrtl").glob("**/*.zig")),
                outf=outf,
                chdir="cxxrtl")
            cr.run()
            shutil.copy(outf, exe_o_path)
        else:
            cmd = [
                "c++",
                # Hard to imagine these flags having any effect.
                *(["-O3"] if args.optimize.opt_code else ["-O0"]),
                *(["-g"] if args.debug else []),
                *cc_o_paths,
                "-o",
                exe_o_path,
            ]
            cr.add_process(cmd,
                infs=cc_o_paths,
                outf=exe_o_path)
            cr.run()

    if not args.compile:
        cmd = [exe_o_path]
        if args.vcd:
            cmd += ["--vcd", args.vcd]
        with logtime(logging.DEBUG, "run"):
            cr.run_cmd(cmd, step="run")


def _make_absolute(path):
    if path.is_absolute():
        try:
            path = path.relative_to(Path.cwd())
        except ValueError:
            raise AssertionError("path must be relative to cwd for builtin-yosys to access it")
    return path



import pytest
from argparse import ArgumentParser

from amaranth import Elaboratable, Module
from amaranth_boards.icebreaker import ICEBreakerPlatform

from niar import Project, build, logging


class FixtureTop(Elaboratable):
    def elaborate(self, platform):
        return Module()


class FixtureProject(Project):
    name = "fixture"
    top = FixtureTop
    targets = [ICEBreakerPlatform]


def test_build_works():
    parser = ArgumentParser()
    build.add_arguments(FixtureProject(), parser)
    args, _argv = parser.parse_known_args()
    args.func(args)



import hashlib
import inspect
import shlex
import subprocess

from .logging import logger

__all__ = ["CompilationUnit", "CommandRunner"]


class CompilationUnit:
    def __init__(self, cmd, *, infs, outf, chdir):
        if inspect.isfunction(cmd):
            self.cmd = cmd
        else:
            self.cmd = [str(el) for el in cmd]
        self.infs  = [str(p) for p in infs]
        self.outf  = str(outf) if outf else None
        self.chdir = chdir

        self.forced = False
        self.digest_path = f"{outf}.dig"

    @property
    def up_to_date(self):
        if self.outf is None:
            return False
        try:
            with open(self.digest_path, "r") as f:
                return f.read() == self.digest_ins_with_cmd()
        except FileNotFoundError:
            return False

    def mark_up_to_date(self):
        if self.outf is None:
            return
        with open(self.digest_path, "w") as f:
            f.write(self.digest_ins_with_cmd())

    def digest_ins_with_cmd(self):
        m = hashlib.sha256()

        def digest_int(i):
            m.update(f"{i:08x}".encode())

        def digest_bytes(b):
            digest_int(len(b))
            m.update(b)

        def digest_str(s):
            digest_bytes(s.encode())

        sorted_in_paths = list(self.infs)
        sorted_in_paths.sort()

        digest_int(len(sorted_in_paths))
        for in_path in sorted_in_paths:
            digest_str(in_path)
            with open(in_path, "rb") as f:
                digest_bytes(f.read())

        if not inspect.isfunction(self.cmd):
            digest_int(len(self.cmd))
            for el in self.cmd:
                digest_str(el)

        return m.hexdigest()


class CommandRunner:
    cus: list[CompilationUnit]

    def __init__(self, *, force=False):
        self.cus = []
        self.force = force

    @property
    def compile_commands(self):
        return {cu.outf: cu.cmd for cu in self.cus}

    def add_process(self, cmd, *, infs, outf, chdir=None):
        self.cus.append(
            CompilationUnit(cmd, infs=infs, outf=outf, chdir=chdir))

    def run(self, step="compile"):
        self.run_cus(self.cus, step)
        self.cus = []

    def run_cmd(self, cmd, *, step="compile", chdir=None):
        cu = CompilationUnit(cmd, infs=[], outf=None, chdir=chdir)
        self.run_cus([cu], step)

    def run_cus(self, cus, step):
        runnables = []
        for cu in cus:
            if cu.up_to_date:
                if self.force:
                    cu.forced = True
                    runnables.append([cu, None])
                else:
                    self.report(cu, skip=True)
            else:
                runnables.append([cu, None])

        for cu_proc in runnables:
            cu = cu_proc[0]
            self.report(cu)
            if inspect.isfunction(cu.cmd):
                cu.cmd()
            else:
                cu_proc[1] = subprocess.Popen(cu.cmd, cwd=cu.chdir)

        failed = []
        for cu, proc in runnables:
            if proc is not None and proc.wait() != 0:
                failed.append(cu)

        if failed:
            logger.error("the following process(es) failed:")
            for cu in failed:
                logger.error(f"  {formatted(cu)}")
            raise RuntimeError(f"failed {step} step")

        for cu, _ in runnables:
            cu.mark_up_to_date()

    def report(self, cu, *, skip=False):
        if cu.forced:
            assert(not skip)
            action = "[force]"
        elif skip:
            action = "[skip] "
        else:
            action = "[run]  "
        logger.info(f"{action} {formatted(cu)}")


def formatted(cu):
    if inspect.isfunction(cu.cmd):
        return cu.cmd.__name__

    cmd = shlex.join(cu.cmd)
    if cu.chdir:
        return f"(in {cu.chdir}/) {cmd}"
    return cmd



from argparse import ArgumentParser

from . import build, cxxrtl
from .cxxrtl_platform import CxxrtlPlatform
from .project import Project

__all__ = ["Project", "cli", "CxxrtlPlatform"]


def cli(np: Project):
    parser = ArgumentParser(prog=np.name)
    subparsers = parser.add_subparsers(required=True)

    build.add_arguments(
        np,
        subparsers.add_parser(
            "build", help="build the design, and optionally program it"
        ),
    )
    if hasattr(np, "cxxrtl_targets"):
        cxxrtl.add_arguments(
            np, subparsers.add_parser("cxxrtl", help="run the C++ simulator tests")
        )

    args = parser.parse_args()
    args.func(args)
